// Main OpenSceneGraph 3D viewer module for DrivePlatform
// Mike Greene, Trideum Corp., July 17, 2007
//  256-704-6113

#include "orientation.h"
#include "mathfunctions.h"

#include <osgProducer/Viewer>
#include <osgProducer/ViewerEventHandler>
#include <Producer/CameraGroup>
#include <osg/Node>
#include <osg/LineSegment>

#include <osg/Notify>
#include <osg/MatrixTransform>
#include <osgDB/Registry>
#include <osgDB/ReadFile>
#include <osgDB/FileUtils>
#include <osg/Geometry>

#include <osgGA/MatrixManipulator>
#include <osg/PositionAttitudeTransform>
#include <osg/Transform>

#include <osg/PolygonStipple>

#include <map>



#define TENA_PLATFORM xp-vc71-d

#include "KeyboardHandler.h"

#include "OSGDefines.h"
#include "xmlParser.h"

// Global for trigerring FFeedback
extern float globalForceMagnitude ;
extern float globalForcePeriod ;
extern bool  bUseCrossCountryDefault;
extern bool  bUseFF ;
extern float globalX;
extern float globalY;
extern bool bDisableFFMagnitude;

extern float globalHeading;
extern float globalTheta;
extern float globalPhi;

#include "joystickff.h"

bool ORBITING = true;

const double MaxAzRate = osg::DegreesToRadians(30.0);	// stored in radians per second - joystick movement
const double MaxElRate = osg::DegreesToRadians(90.0);	// stored in radians per second - joystick movement
const double MaxVel = (25.0 * 0.447);				    // maximum units (distance) per second - joystick movement
double jsX, jsY;       // joytick vals

// These are the vars filled by TENA output and used to drive the 3D display
double _positionX = 0, _positionY = 0, _positionZ = 0, _heading = 0, _theta = 0, _phi = 0, 
	_elevation = 0, _velocity = 0, _xJitterOffset = 0, _yJitterOffset = 0, _zJitterOffset = 0;
int _UID = 100;

// Global OSG/Windows variables used amoung methods
HWND myOSGHwnd;
std::string sRenderWindowName;
Producer::Camera* mainCamera = new Producer::Camera();  // OSG Camera object
osgDB::FilePathList pathList;// = new osgDB::FilePathList;
osg::PositionAttitudeTransform *followerPAT = new osg::PositionAttitudeTransform();
double radToDeg = osg::RadiansToDegrees(1.0);

void SetOSGWindowViews(osgProducer::Viewer viewer);

// Default view type for top and bottom windows
short globalTopViewType    = COCKPIT;
short globalBottomViewType = GODSEYE;
int   globalInputDevice      = RUMBLEPAD;

char ts[] = "                                                      ";
//XML input terrain and vehicle model info
std::string sXMLTerrainFileName;
std::string sXMLTerrainPathName;

std::string sXMLVehicleFileName;
std::string sXMLVehiclePathName;

double xOffsetToZero, yOffsetToZero, zOffsetToZero; // read in from XML file to position (0,0,0) on terrain
void ReadInXMLPolygonInfo();

void * StartJoystickProcessorThread(void *data);

osg::Quat gViewquat;

// Routine to either circle on terrain, or do terrain clamping, depending on value of
//  boolean ORBITING
// Mike Greene - Trideum   July 16, 2007
// (256) 704-6113
class circleAimlessly : public osg::NodeCallback
{
public:
   circleAimlessly(osg::Node* pTerrain): angle(0.0), pTerrainNode(pTerrain)
   {
   }

   virtual void operator () (osg::Node * node, osg::NodeVisitor* nv)
   {
      osg::PositionAttitudeTransform* pat = dynamic_cast <osg::PositionAttitudeTransform*> (node);
      if (pat != NULL)
      {
		  if(ORBITING)
		  {
			angle += osg::DegreesToRadians(1.0f); 
			//pat->setPosition( osg::Vec3(cosf(angle)*100.0f, sinf(angle)*100.0f, 20.0f));
			//pat->setPosition( osg::Vec3(cosf(angle)*100.0f-XOFFSET_MOUT0, sinf(angle)*100.0f-YOFFSET_MOUT0, 20.0f-ZOFFSET_MOUT0));
			pat->setPosition( osg::Vec3(cosf(angle)*100.0f-xOffsetToZero, 
				              sinf(angle)*100.0f-yOffsetToZero, 20.0f-zOffsetToZero));
			
			pat->setPosition( osg::Vec3d(pat->getPosition().x() + _xJitterOffset, pat->getPosition().y() + _yJitterOffset, pat->getPosition().z() + _zJitterOffset)); 
			
			pat->setAttitude( osg::Quat(angle, osg::Vec3(0,0,1)) );
		  
		    Orientation::Quaternian_to_Euler(pat->getAttitude().w(), pat->getAttitude().x(), pat->getAttitude().y(), pat->getAttitude().z(), _heading, _theta, _phi);
			
		  	_positionX = pat->getPosition().x();
			_positionY = pat->getPosition().y();
			_positionZ = pat->getPosition().z();		  

			_velocity = 10;
		  }
		  else
		  {
			// Mike's terrain clamping code
			// use line segment to project into terrain to determine intersection point
			osg::LineSegment* tankLocationSegment = new osg::LineSegment();
			tankLocationSegment->set(
				osg::Vec3(_positionX, _positionY,  999) ,
				osg::Vec3(_positionX, _positionY, -999) );

			osgUtil::IntersectVisitor findTankElevationVisitor;
			findTankElevationVisitor.addLineSegment(tankLocationSegment);
			pTerrainNode->accept(findTankElevationVisitor);

			osgUtil::IntersectVisitor::HitList tankElevationLocatorHits;
			// if intersection, save hit list (might be more than 1)
			tankElevationLocatorHits =
				findTankElevationVisitor.getHitList(tankLocationSegment);
			
			// outside terrain bounds
			if ( tankElevationLocatorHits.empty() )
			{
				MessageBox(NULL, " couldn't place tank on terrain" ,"",MB_OK);
				pat->setPosition( osg::Vec3d(_positionX + _xJitterOffset, _positionY + _yJitterOffset, _positionZ + _zJitterOffset) );
				pat->setAttitude(osg::Quat(_heading,osg::Vec3(0,0,1)));
			}
			// got some hits with intersection
			else 
			{
				osgUtil::Hit heightTestResults = tankElevationLocatorHits.front();
				osg::Vec3 terrainHeight =
					heightTestResults.getWorldIntersectPoint();
				osg::Vec3d np = heightTestResults.getLocalIntersectNormal();
				
				pat->setPosition( osg::Vec3d(terrainHeight.x() + _xJitterOffset, terrainHeight.y() + _yJitterOffset, terrainHeight.z() + _zJitterOffset) );

				_positionX = terrainHeight.x();
				_positionY = terrainHeight.y();
				_positionZ = terrainHeight.z();	

			// constant up vector 
			osg::Vec3d up ( 0.0f, 0.0f, 1.0f ); 
			
			// make the rotation from the up vector to the normal 
			 
			gViewquat.makeRotate ( up, np ); 
			// mg retrieve the terrain following Eulers
			double mpsi, mtheta, mphi;

			mpsi = 0.0;
			mtheta = 0.0;
			mphi = 0.0;
			Orientation::Quaternian_to_Euler(gViewquat.w(),gViewquat.x(),gViewquat.y(),gViewquat.z(),
					mpsi,mtheta,mphi);
			mpsi   += _heading;
			mtheta += _theta;
			mphi   += _phi;

			globalHeading = mpsi;
			globalTheta = mtheta;
			globalPhi = mphi;

			// set output variables to include terrain following angles
			/*_heading = mpsi;
			_theta   = mtheta;
			_phi     = mphi;*/

			Orientation::Euler_to_Quaternian(mpsi,mtheta,mphi,
				gViewquat.w(),gViewquat.x(),gViewquat.y(),gViewquat.z());
	 // update tank's orientation based on calculated pitch and roll
			pat->setAttitude(gViewquat);

			
				
				//pat->setAttitude(osg::Quat(_heading,osg::Vec3(0,0,1)));
			 }
		}

		traverse(node, nv);		
      }
   }
private:
      float angle;
	  double w, x, y, z;
	  osg::Node* pTerrainNode;
};

// class to allow access to matrix that represents accumlation of 
//  matrices above specified node of scene graph.
struct updateAccumulatedMatrix : public osg::NodeCallback
{
   virtual void operator()(osg::Node* node, osg::NodeVisitor* nv)
   {
	  matrix = osg::computeWorldToLocal(nv->getNodePath() );
	  traverse(node,nv);
   }
   osg::Matrix matrix;
};

struct transformAccumulator
{
   transformAccumulator();
   bool attachToGroup(osg::Group* g);
   osg::Matrix getMatrix();
protected:
   osg::ref_ptr<osg::Group> parent;
   osg::Node* node;
   updateAccumulatedMatrix* mpcb;
};

osg::Matrix transformAccumulator::getMatrix()
{
   return mpcb->matrix;
}
transformAccumulator::transformAccumulator()
{
   parent = NULL;
   node = new  osg::Node;
   mpcb = new updateAccumulatedMatrix();
   node->setUpdateCallback(mpcb);
}

bool transformAccumulator::attachToGroup(osg::Group* g)
{
   bool success = false;
   if (parent != NULL)
   {
      int n = parent->getNumChildren();
      for (int i = 0; i < n; i++)
      {
         if (node == parent->getChild(i) )
         {
            parent->removeChild(i,1);
            success = true;
         }
      }
      if (! success)
      {
         return success;
      }
   }
   g->addChild(node);
   return true;
}

class followNodeMatrixManipulator : public osgGA::MatrixManipulator
{
public:
   followNodeMatrixManipulator( transformAccumulator* ta) {worldCoordinatesOfNode = ta; theMatrix = osg::Matrixd::identity();}
   bool handle (const osgGA::GUIEventAdapter&ea, osgGA::GUIActionAdapter&aa);
   void updateTheMatrix();
   virtual void setByMatrix(const osg::Matrixd& mat); 
   virtual void setByInverseMatrix(const osg::Matrixd&mat);
   virtual osg::Matrixd getInverseMatrix() const;
   virtual osg::Matrixd getMatrix() const;
protected:
   ~followNodeMatrixManipulator() {}
   transformAccumulator* worldCoordinatesOfNode;
   osg::Matrixd theMatrix;
};

void followNodeMatrixManipulator::setByMatrix(const osg::Matrixd& mat)
{
   theMatrix = mat;
}
void followNodeMatrixManipulator::setByInverseMatrix(const osg::Matrixd& mat)
{
   theMatrix = mat.inverse(mat);
}
void followNodeMatrixManipulator::updateTheMatrix()
{
   theMatrix = worldCoordinatesOfNode->getMatrix();
}
osg::Matrixd followNodeMatrixManipulator::getInverseMatrix() const
{
   osg::Matrixd m;
   m = theMatrix * osg::Matrixd::rotate(-M_PI/2.0, osg::Vec3(1,0,0) );
   return m;
}

osg::Matrixd followNodeMatrixManipulator::getMatrix() const
{
   return theMatrix;
}

bool followNodeMatrixManipulator::handle(const osgGA::GUIEventAdapter&ea, osgGA::GUIActionAdapter&aa)
{
   switch(ea.getEventType())
   {
   case (osgGA::GUIEventAdapter::FRAME):
      {
         updateTheMatrix();
         return false;
      }
   }
   return false;
}

void SetOrbiting()
{
	ORBITING = !ORBITING;
}

// Joystick input processing routine
// Mike Greene , Trideum Corp., 05-02-07
//
void 
ProcessPositionOrientation(double frameRate, double dx, double dy, double nullTolerance)
             //vtBoolean btn1, vtBoolean btn2, vtBoolean btn3)
{
   // dx and dy span -1.0 to 1.0 
   // evaluate null region
   if (fabs(dx) < nullTolerance)
   {
      dx = 0.0f;
   }
   if (fabs(dy) < nullTolerance)
   {
      dy = 0.0f;
   }

   _heading -= (dx * MaxAzRate) / frameRate;
   _heading = fmod(_heading,osg::DegreesToRadians(360.0));
   
   _velocity = (-dy * MaxVel);   
   _elevation = 0.0;
   
	// compute new position based on attitude and velocity
	_positionZ += (_velocity / frameRate) * sin(_elevation);
	double grndDist = (_velocity / frameRate) * cos(_elevation);
	_positionY += grndDist * cos(_heading);
	_positionX -= grndDist * sin(_heading);
 
}

// Routine to set up two camera views along with their screen real estate and aspect ratios
// Mike Greene - Trideum Corp.    July 17, 2007
// (256) 704-6113
Producer::CameraConfig* setupCameras()
{
   Producer::CameraConfig* myCameraConfig = new Producer::CameraConfig();
   Producer::Camera* tankFollowerCamera = new Producer::Camera();
   
// Bottom camera
   mainCamera->setProjectionRectangle(0.1f, 0.9f, 0.0f, 0.5f);
   mainCamera->setShareLens(false);
   mainCamera->getLens()->setAutoAspect(true);
   mainCamera->setClearColor(0.8f,0.8f,0.8f,0.0f);
   
// Top Camera
   tankFollowerCamera->setProjectionRectangle(0.1f, 0.9f, 0.55f, 0.95f);
   tankFollowerCamera->setShareLens(false);
   tankFollowerCamera->getLens()->setAutoAspect(true);
   tankFollowerCamera->setClearColor(0.8f,0.8f,0.8f,0.0f);
   

   myCameraConfig->addCamera("tankFollowerCamera",tankFollowerCamera);
   myCameraConfig->addCamera("mainCamera",mainCamera);

   Producer::RenderSurface* rsOne = mainCamera->getRenderSurface();
   rsOne = mainCamera->getRenderSurface();
   tankFollowerCamera->setRenderSurface( rsOne );
   rsOne->setWindowName("TA2");
   //rsOne->setWindow(GetSafeHwnd());
   //rsOne->makeCurrent();
   // set up window size and location on screen as defined by these global #defines
   rsOne->setWindowRectangle(OSGWINDOW_XMIN,OSGWINDOW_YMIN,OSGWINDOW_XMAX,OSGWINDOW_YMAX);
   sRenderWindowName = rsOne->getWindowName();

   return myCameraConfig;
}

void OSG_Module(void)
{
   osg::Group* rootNode = new osg::Group;

   // read in parh and model initialization from xml file
   pathList = osgDB::getDataFilePathList();
   ReadInXMLPolygonInfo();
   //MessageBox(NULL,"After XML read","",MB_OK);
  
   osgDB::setDataFilePathList(pathList);

   // set terrain model based on info read in from XML file
   osg::Node* terrainNode = osgDB::readNodeFile(sXMLTerrainFileName);
   
   if (!terrainNode)
   {
      MessageBox(NULL, " no terrain! " ,"Error!",MB_OK);
      return ;
   }
   rootNode->addChild(terrainNode);

   // set UUT (Platform) model based on info read in from XML file
   osg::Node* tankNode = osgDB::readNodeFile(sXMLVehicleFileName);
   if( ! tankNode)
   {
       MessageBox(NULL, " no UUT! " ,"Error!",MB_OK);
      return ;
   }
	
   // set up two cameras and add scene data to viewer
   Producer::CameraConfig* myCameraConfig = setupCameras();
   osgProducer::Viewer viewer(myCameraConfig);
   viewer.setSceneData( rootNode );
   viewer.setUpViewer(osgProducer::Viewer::STANDARD_SETTINGS);

   // Create and set up a transform for updating the tank's position and attitude
   osg::PositionAttitudeTransform* tankPAT = new osg::PositionAttitudeTransform();

   tankPAT->setUpdateCallback(new circleAimlessly(terrainNode));
   rootNode->addChild(tankPAT);
   tankPAT->addChild(tankNode);
   
   // Declare and set up a transform (iffset from tank center) to 'follow' the tank node.
   
   followerPAT->setPosition( osg::Vec3(0,-15,4) ); 
   followerPAT->setAttitude( osg::Quat( osg::DegreesToRadians(-10.0f) , 
      osg::Vec3(1,0,0) ));

   // Add this as a child of the tank's transform
   tankPAT->addChild(followerPAT);

   // create the windows and run the threads.
   viewer.realize();

   transformAccumulator* tankWorldCoords = new transformAccumulator();
   tankWorldCoords->attachToGroup(followerPAT);
   followNodeMatrixManipulator* followTank = new followNodeMatrixManipulator(tankWorldCoords);
   osgGA::KeySwitchMatrixManipulator *ksmm = viewer.getKeySwitchMatrixManipulator();
   if (!ksmm)
      return ;
   // add the tank follower matrix manipulator. Selecting the 'f' key 
   // with switch the viewer to this manipulator.
   ksmm->addMatrixManipulator('m',"tankFollower",followTank);
   ksmm->selectMatrixManipulator(ksmm->getNumMatrixManipulators()-1);

   // the "O" key will swith between canned circular motion and usage of TENA inputs
   keyboardEventHandler* key_o = new keyboardEventHandler();
   key_o->addFunction('o',SetOrbiting);
   viewer.getEventHandlerList().push_front(key_o);

   myOSGHwnd = FindWindow(NULL,"TA2");

   if( FAILED( InitDirectInput( myOSGHwnd ) ) )
   {
		MessageBox(NULL,  "Error Initializing FFBack - Unable to use FFBack" ,"",MB_OK);
   }

   if( FAILED( InitDirectInputj( myOSGHwnd ) ) )
   {
		MessageBox(NULL, "Error Initializing DirectInput - Unable to use joystick", "",MB_OK);
   }

   // main OSG timing loop - updates scene and renders to screen
   while( !viewer.done() )
   {
      // wait for all cull and draw threads to complete.
      viewer.sync();

      // update the scene by traversing it with the the update visitor which will
      // call all node update callbacks and animations.
      viewer.update();

	  // based on user input, set up views
	  SetOSGWindowViews(viewer);

	  UpdateInputState(jsX, jsY);  // get joystick X & Y
		
	  if(!ORBITING)
			ProcessPositionOrientation(2.55, jsX, jsY, 0.1);
	  //StartFF();

	  // fire off the cull and draw traversals of the scene.
      viewer.frame();
   }

   // wait for all cull and draw threads to complete before exit.
   viewer.sync();

   return ;

}

// Routine to read in OSG terrain and UUT model and path information
// Mike Greene  -  Trideum Corp.   July 17,2007
//  (256)  704-6113
void ReadInXMLPolygonInfo()
{
	 // this opens and parses the XML file:
	XMLNode xMainNode=XMLNode::openFileHelper("viewerconfig.xml","MGXML");

    XMLNode xNode=xMainNode.getChildNode("terrainfile");
	sXMLTerrainFileName = std::string(xNode.getText());
	//MessageBox(NULL,xNode.getText(),"TerrainFileName",MB_OK);

	xNode=xMainNode.getChildNode("vehiclefile");
	sXMLVehicleFileName = std::string(xNode.getText());
	//MessageBox(NULL,xNode.getText(),"VehicleFileName",MB_OK);

	xNode=xMainNode.getChildNode("xoffset");
	xOffsetToZero = atof(xNode.getText());
	//MessageBox(NULL,xNode.getText(),"VehicleFileName",MB_OK);

	xNode=xMainNode.getChildNode("yoffset");
	yOffsetToZero = atof(xNode.getText());
	//MessageBox(NULL,xNode.getText(),"VehicleFileName",MB_OK);

	xNode=xMainNode.getChildNode("zoffset");
	zOffsetToZero = atof(xNode.getText());
	//MessageBox(NULL,xNode.getText(),"VehicleFileName",MB_OK);
	

	// find number of "modelpath" tags
	int np=xMainNode.nChildNode("modelpath");
	sprintf(ts,"%d = numpaths",np);
	//MessageBox(NULL,ts,"",MB_OK);
	

	int j, polyIterator=0;
	for (j=0 ; j < np; j++)
	{
		// this gets the number data from the "path" tags and pushes them onto the search list:
		
		xNode = xMainNode.getChildNode("modelpath",&polyIterator).getChildNode("path");
		//MessageBox(NULL,xNode.getText(),"Path = ",MB_OK);
		pathList.push_back(xNode.getText());	
	}
    
}

// Routine to set up user defined views for top and bottom windows
// Mike Greene - Trideum Corp.   July 17, 2007
//  (256) 704-6113
void SetOSGWindowViews(osgProducer::Viewer viewer)
{
	// local vars used to manipulate slider values
   double rotateAngle = 0.0;
   double panAng  = 0.0;
   double tiltAng = 0.0;
   Producer::Matrix m2;
   Producer::Matrix m;

   // based on which radio button is selected in the MFC gui, adjust view as needed
	switch (globalTopViewType) {
		case GODSEYE:
			viewer.getCameraConfig()->findCamera("tankFollowerCamera")->setViewByLookat(_positionX-5,_positionY-5,GODS_EYE_HEIGHT,
		            _positionX,_positionY,_positionZ,
					0.,1.0,0.0);

			//viewer.getCameraConfig()->findCamera("tankFollowerCamera")->setLensPerspective(60,
			//	            30,1.,10000.);
			break;
		case HOVER:
            followerPAT->setPosition( osg::Vec3(0,.1,20) ); //was 0, -22, 4
			followerPAT->setAttitude( osg::Quat( osg::DegreesToRadians(-90.0f) , 
				osg::Vec3(1,0,0) ));
			//viewer.getCameraConfig()->findCamera("tankFollowerCamera")->setLensPerspective(60,
			//	            30,1.,10000.);
			break;
		case COCKPIT:
            followerPAT->setPosition( osg::Vec3(0,5,1.5) ); //was 0, -22, 4
			followerPAT->setAttitude( osg::Quat( osg::DegreesToRadians(1.0f) , 
				osg::Vec3(1,0,0) ));
			//viewer.getCameraConfig()->findCamera("tankFollowerCamera")->setLensPerspective(globalOSGHFOVValue,
			//	            globalOSGVFOVValue,1.,10000.);
			break;

		case FOLLOW:
            followerPAT->setPosition( osg::Vec3(0,-20,4) ); //was 0, -22, 4
			followerPAT->setAttitude( osg::Quat( osg::DegreesToRadians(-10.0f) , 
				osg::Vec3(1,0,0) ));
			//viewer.getCameraConfig()->findCamera("tankFollowerCamera")->setLensPerspective(60,
			//	            30,1.,10000.);
			break;

		
	}

	// for now, bottom view can onlt be God's Eye
	switch (globalBottomViewType) {
		case GODSEYE:
			viewer.getCameraConfig()->findCamera("mainCamera")->setViewByLookat(_positionX-5,_positionY-5,GODS_EYE_HEIGHT,
		            _positionX,_positionY,_positionZ,
					0.,1.0,0.0);
			break;

	}
}

