/////////////////////////////////////////////////////////////////// -*- C++ -*-
/*! \file RTTC/EnvironmentalEffects/DisturbanceHistoryMessage/BasicImpl/MessageCallbackFactoryImpl.h
 * \version 5.2.1
 *
 * \brief Contains the declaration of and documentation for
 * RTTC::EnvironmentalEffects::DisturbanceHistoryMessage::BasicImpl::MessageCallbackFactoryImpl.
 *
 * This file contains the declaration of and documentation for the
 * RTTC::EnvironmentalEffects::DisturbanceHistoryMessage::BasicImpl::MessageCallbackFactoryImpl class.
 */

#ifndef RTTC_EnvironmentalEffects_DisturbanceHistoryMessage_BasicImpl_MessageCallbackFactoryImpl_h
#define RTTC_EnvironmentalEffects_DisturbanceHistoryMessage_BasicImpl_MessageCallbackFactoryImpl_h

#include <DCT/System/config.h>
#include <basicImpl_RTTC_EnvironmentalEffects_v1_EXPORT.h>
#include <memory>
#include <RTTC/EnvironmentalEffects/DisturbanceHistoryMessage/MessageCallbackFactory.h>
#include <DCT/Utils/SmartPtr.h>
#include "MessageCallbackImpl.h"
#include "CallbackInfo.h"

// The example applications provided with the TENA Middleware utilize
// namespaces to separate class declarations and definitions.  This is not
// required, but is a recommended C++ practice.
namespace RTTC
{
  namespace EnvironmentalEffects
  {
    namespace DisturbanceHistoryMessage
    {
      namespace BasicImpl
      {
        class CallbackInfo;
        typedef DCT::Utils::SmartPtr<CallbackInfo> CallbackInfoPtr;
        /*!\class MessageCallbackFactoryImpl MessageCallbackFactoryImpl.h \
           RTTC/EnvironmentalEffects/DisturbanceHistoryMessage/BasicImpl/MessageCallbackFactoryImpl.h
         *
         * \brief Example of a concrete implementation of an \
           RTTC::EnvironmentalEffects::DisturbanceHistoryMessageCallbackFactory.
         *
         * This is a user-defined class that allows the TENA Middleware to
         * create Callback objects when a RTTC::EnvironmentalEffects::DisturbanceHistoryMessage::Message
         * is transmitted via the TENA Message Service.
         *
         * This class is provided to serve as an example for the TENA Middleware
         * user.
         */
        class basicImpl_RTTC_EnvironmentalEffects_v1_EXPORT MessageCallbackFactoryImpl
          : public virtual RTTC::EnvironmentalEffects::DisturbanceHistoryMessage::MessageCallbackFactory
        {
        public:

          MessageCallbackFactoryImpl( CallbackInfoPtr const & );

          virtual
          ~MessageCallbackFactoryImpl();

          //! Create a Callback and return it to the TENA Middleware.
          /*!
            \param[in] pMessage A Pointer for the RTTC::EnvironmentalEffects::DisturbanceHistoryMessage::Message
            that was transmitted via the TENA Message Service

            \return An object that implements the TENA::Middleware::Callback::Callback
            interface.  Ownership of the memory is passed to the invoker.
          */
          std::auto_ptr< TENA::Middleware::Callback::Callback >
          create( RTTC::EnvironmentalEffects::DisturbanceHistoryMessage::Pointer const & pMessage );

        private:
          //! Default copy construction is disallowed.
          MessageCallbackFactoryImpl(
            MessageCallbackFactoryImpl const & rhs );
          //! Default copy assignment is disallowed.
          MessageCallbackFactoryImpl & operator=(
            MessageCallbackFactoryImpl const & rhs );

          CallbackInfoPtr _pCallbackInfo;
        };
      } // End of namespace BasicImpl
    } // End of namespace DisturbanceHistoryMessage
  } // End of namespace EnvironmentalEffects
} // End of namespace RTTC

#endif // ifndef RTTC_EnvironmentalEffects_DisturbanceHistoryMessage_BasicImpl_MessageCallbackFactoryImpl_h
